#!/usr/bin/perl
# Generate a Logarithmic or Archimedes Spiral Antenna

my $VERSION = '2020-10-08';
# TO-DO LIST: Nothing.

print "\nKY8D_Spiral_Calculator.pl version $VERSION";

use Getopt::Long;
use Math::Trig ('tan', 'deg2rad', ':radial');

my $help = my $iter = my $lgth_arm = $lgth_sib = my $flip = my $spin = my $f_md = 0;
my ($c, $pi, $ohm, $div, $fat, $path) = (299.792458, 3.14159, 900, 6, 1, './');

GetOptions(
	'ohm=f'  => \$ohm,     # Expected impedance
	'deg=f'  => \my $deg,  # Angle of cone
	'div=i'  => \$div,     # Divisions of circle
	'dia=s'  => \my $dia,  # Diameter of wire as mm.
	'tau=f'  => \my $tau,  # Expansion factor.
	'zee=f'  => \my $zee,  # 1st Z division
	'seg=i'  => \my $seg,  # Segments per wire.
	'fat=f'  => \$fat,     # Two wires per leg. Spread: 1, 2, 3.
	'hgt=f'  => \my $hgt,  # Height above ground
	'f_hi=f' => \my $f_hi, # Frequency, highest
	'f_md=f' => \$f_md,    # Frequency, for calculation
	'f_lo=f' => \my $f_lo, # Frequency, lowest
	'path=s' => \$path,    # Output file pathpath
	'iter=s' => \$iter,    # Make plural files
	'0th=f'  => \my $iter_0th, 
	'nth=f'  => \my $iter_Nth, 
	'stp=f'  => \my $iter_step, 
	'stub'   => \$stub,    # Short conical spirals far ends together.
	'spin=f' => \$spin,    # Rotate array by degrees.
	'flip'   => \$flip,    # Flip array to vertical.
	'help'   => \$help,    # Give help
) or die "\nInvalid options passed to $0. Call again with arg --help\n";

# Translate from full_circle to radius_divison
# So that can change --div and keep same --tau
$tau_360 = $tau; # Preserve for iterating on --div 
$tau = $tau**(1/$div); # Adapt to value of --div

# ANT files want a plot freq. Make up if user didn't give.
unless ($f_md) { $f_md = ($f_hi + $f_lo)/2 }
unless ($seg) { $seg = 1 }

# Set parameter limits.
my $rho = $c/$f_hi/4; # Initial rho = radius of highest frequency.
my $max = $c/$f_lo/4; # Max rho = radius of lowest frequency.

# XYZ of spiral arm ends.
my @end_1; my @end_1_sib;
my @end_2; my @end_2_sib;
my $lowest_z = 999; # Flag to track if above ground.

sub sib {
	my $r = shift;
	# return 1 + ($r-1)/2;
	return $r;
}

# Display a title and list. A t-shooting aid.
# Call as blab('Line 113', foo, bar, this, that, other);
sub blab {
   my $header = shift;
   print "\n$header:\n\t"  . join("\n\t", @_) . "\n";
}

if ($help) {
   my $s = "\n\nDesign tool for planar and conical spiral antennas.";
   $s .= "Output file types: *.nec, *.ant, and *.txt (Rhino 6 CAD macro).";
   $s .= "\nRequired Args: --f_hi 40 --f_lo 12 --div 6 --dia 2 --tau 1.05";
   $s .= "\n\t f_hi = MHz of highest freqeuncy.";
   $s .= "\n\t f_md = MHz named for plotting within the file.";
   $s .= "\n\t f_lo = MHz of lowests frequency.";
   $s .= "\n\t seg = segments per wire (1 <= N)";
   $s .= "\n\t div = divisions of circle (N = 4,6,8,10,12). Default = 6.";
   $s .= "\n\t tau = expansion factor. If 1 < N, multiplies. If N <= 1, adds N meters.";
   $s .= "\n\t deg = degrees for the cone (0 <= N <= 60)";
   $s .= "\n\t hgt = m feed point height-above-ground (0.0 <= N)";
   $s .= "\n\t zee = mm spacing in Z for conicals (0.0 <= N)";
   $s .= "\n\t dia = mm or #awg diameter of wire (0.0 < N)";
   $s .= "\n\t fat = fatten spirals (1 <= N <= 3). Default = 1";
   $s .= "\n\t ohm = Z Ohms expected (if different from $ohm)";
   $s .= "\n\t iter = output plural files: any arg above.";
   $s .= "\n\t\t --0th = iter start value.";
   $s .= "\n\t\t --nth = iter finish value.";
   $s .= "\n\t\t --stp = iter step value.";
   $s .= "\n\t path = output file path (c:/foo/bar).";
   $s .= "\n\t stub = short spiral arms at end.";
   $s .= "\n\t spin = rotate array 90 degs in same plane.";
   $s .= "\n\t flip = rotat array into vertical.";
   $s .= "\n\t help = show this info";
   $s .= "\nPerl script by KY8D. No rights reserved.";
   print "$s\n";
}

my $zee_sib = $zee; # Not sure about using sib on this yet.

# A couple of reminders.
if ( $deg == 0 && $zee != 0) {
	print "\nNote! Calling --zee without --deg is useless.\n";
} elsif ( $deg != 0 && $zee == 0) {
	print "\nNote! Calling --deg without --zee is useless.\n";
}

$dia = convert_diameter($dia); # meters from mm or AWG

my %iter_start = (
	'max' => $max,
	'rho' => $rho,	
	'rho_sib' => $rho,	
	'tau' => $tau,		
	'add' => $add,		
	'add_sib' => $add,			
	'zee' => $zee,		
	'zee_sib' => $zee,
);

my ($cnt, $ang, $theta, $nec, $ant, $rhino);
my @xyz_fm; my @xyz_fm_sib;
my @xyz_to; my @xyz_to_sib;
my @main; my @main_sib;
my @anti; my @anti_sib;
	
sub init_ant {
	my ($string, $iter) = @_;
	$cnt = 0; # Wire number.
	$ang = 0; # Protractor angle of last wire end on XY plane.
	$theta = 0;
	$nec = "CM Generted by $0 version $VERSION";
	$ant  = "Spiral Antenna";
	$rhino  = "; Spiral Antenna";
	if ($string ne '') {
		$nec .= "\nCM Iteration: $string = $iter";
		$ant  .= ", Iteration $string = $iter";
		$rhino  .= ", Iteration $string = $iter";		
	}
	if ($lowest_z > 0 ) {		
		$ant .= "\nSommerfeld-Norton Ground";
	} else {
		$ant .= "\nFree Space";		
	}
	$ant .= "\nFrequency $f_md MHz Zin=$ohm Segs=" . int($seg/2) . ', Copper wires, Meters;';
	$ant .= "\nSource: center;";
	
	# Spirals build from off feed wire ends
	@xyz_fm = @xyz_fm_sib = spin(-$rho, 0, 0, $spin); # XYZ format
	@xyz_to = @xyz_to_sib = spin( $rho, 0, 0, $spin); # XYZ format
	
	# Spirals begin at ends of feed wire.
	@main = @main_sib = ($xyz_to[0], $xyz_to[1], $xyz_to[2]);
	@anti = @anti_sib = ($xyz_fm[0], $xyz_fm[1], $xyz_fm[2]);
}

# Increment cylindrical degrees to next guy line.
sub inc_theta {
	$theta += deg2rad(360/$div);
	if ($theta < 1 && $theta > 359) { $theta = 0 }
}


# Rotate by degrees on XY plane.
# Do BEFORE flip.
sub spin {
	my ($x, $y, $z_1, $degs) = @_;
	my ($rho, $theta, $z_2) = cartesian_to_cylindrical($x, $y, $z_1);
	my @spun = cylindrical_to_cartesian($rho, $theta + deg2rad($degs), $z_2);
	# *.ant files cannot abide exponential notation
	for (0,1,2) {$spun[$_] = sprintf('%0.5f', $spun[$_])}
	return @spun;
}

sub wire {
	my @fm = @xyz_fm; my @fm_sib = @xyz_fm_sib;
	my @to = @xyz_to; my @to_sib = @xyz_to_sib;
	@end_1 = @xyz_fm = @xyz_to;
	@end_1_sib = @xyz_fm_sib = @xyz_to_sib;
	inc_theta();
	
	$zee *= ($tau-1)/$div + 1; # Might remain zero, if a planar spiral.
	
	# Manage expansion, whether planar or conical.
	my $rho_sib;
	if (abs($zee) == 0) {
		# Is a planar spiral.		
		if ($tau > 1) {
			# Good split on geometry with **3, but impedance worse.
			$rho_sib = $rho * $tau**$fat;
			# For regular two wires.
			$rho *= $tau;			
		} else {
			$rho += $tau;			
		}		
	} else {
		# Is a conical spiral.
		$rho += abs($zee)* tan(deg2rad($deg));	# conical spiral		
	}
	# Point down with pos Z, point up with neg Z
	@xyz_to = cylindrical_to_cartesian($rho, $theta, $xyz_fm[2] + $zee); 
	@xyz_to_sib = cylindrical_to_cartesian($rho_sib, $theta, $xyz_fm_sib[2] + $zee);
	for my $i (0,1,2) {
		$xyz_to[$i] = sprintf('%.5f', $xyz_to[$i] );
		$xyz_to_sib[$i] = sprintf('%.5f', $xyz_to_sib[$i] );
	}
	if ($spin) {
		@xyz_fm = spin(@xyz_fm, $spin);
		@xyz_to = spin(@xyz_to, $spin);
	}
	push @main, ( $xyz_to[0],  $xyz_to[1], $xyz_to[2]);
	push @anti, (-$xyz_to[0], -$xyz_to[1], $xyz_to[2]);
	push @main_sib, ( $xyz_to_sib[0],  $xyz_to_sib[1], $xyz_to_sib[2]);
	push @anti_sib, (-$xyz_to_sib[0], -$xyz_to_sib[1], $xyz_to_sib[2]);
	$lgth_arm += accum_lgth(\@main);
	$lgth_sib += accum_lgth(\@main_sib);
}

sub accum_lgth {
	my ($aref) = @_; # Has XYZ-from, XYZ-to values at end.
	my $dx = $aref->[-1] - $aref->[-4];
	my $dy = $aref->[-2] - $aref->[-5];
	my $dz = $aref->[-3] - $aref->[-6];
	return sqrt($dx**2+$dy**2+$dz**2);
}

# Short spiral ends if --stub flagged.
sub stub {
	if ($stub) {
		if ($deg != 0) {
			$nec .= qq{\nGW $cnt $seg } . join(' ', @end_1, @end_2, $dia/2);
			$ant .= qq{\n$cnt } . join(',', @end_1) . ' ' . join(',', @end_2, $dia);
			$rhino .= qq{\n_Line } . join(',', @end_1) . ' ' . join(',', @end_2);	
		} else {
			print "\nNote! The -stub option for conical spirals only. Ignored.";
		}
	}
}

sub convert_diameter {
	my $mm = shift; # Might be '#14' for 14 AWG
	if ($mm eq '#22') {$mm = 0.6438}
	if ($mm eq '#21') {$mm = 0.7229}
	if ($mm eq '#20') {$mm = 0.8118}
	if ($mm eq '#19') {$mm = 0.9116}
	if ($mm eq '#18') {$mm = 1.0237}
	if ($mm eq '#17') {$mm = 1.1495}
	if ($mm eq '#16') {$mm = 1.2908}
	if ($mm eq '#15') {$mm = 1.4495}
	if ($mm eq '#14') {$mm = 1.6277}
	if ($mm eq '#13') {$mm = 1.8278}
	if ($mm eq '#12') {$mm = 2.0525}
	if ($mm eq '#11') {$mm = 2.3048}
	if ($mm eq '#10') {$mm = 2.5882}
	if ($mm eq '#9')   {$mm = 2.9064}
	if ($mm eq '#8')   {$mm = 3.2636}
	if ($mm =~ m/#/) { die("\nOops! Range: 8 >= AWG >= 22") }
	if ($mm =~ m/mm/) { print "\nNote: Metric assumed if no # for AWG."}
	$mm =~ s/mm//; # User may have typed 'mm'
	return $mm/1000;
}

my @feed;

sub build_feed_wire {
	++$cnt;
	@feed = (@main[@_], @anti[@_]);
}

sub fatten_feed_wire {
	@feed = ($feed[0]/2, @feed[1,2], $feed[3]/2, @feed[4,5]);	
}

# Feed wire needs odd segments
sub feed {
	my ($feed_segs, $ex_seg) = (1,1);
	if ($seg != 1) {
		$feed_segs = 2 * int($seg/2) + 1;
		$ex_seg = int($seg/2) + 1;
	}
	return ($feed_segs, $ex_seg); 
}

sub write_feed_wire {
	my @feed_segs = feed();
	my @ends = @feed[@_];
	if ($spin && $flip) {
		@ends = @ends[2,1,0,5,4,3];
	}
	$ends[2] += $hgt; # Adjust height.
	$ends[5] += $hgt;
	$nec .= qq{\nCM First wire is feed line.};
	$nec .= qq{\nGW $cnt $feed_segs[0] } . join(' ', @ends, $dia/2);
	$ant .= qq{\n$cnt } . join(' ', @ends, $dia);
	$rhino .= qq{\n; First wire is feed line.};	
	$rhino .= qq{\n_Line } . join(',', @ends[@_[0,1,2]]) . ' ' . join(',', @ends[@_[3,4,5]]);		
}

sub write_spiral { 
	my $aref = shift; # Six values remain on @_
	my @ends = @$aref[@_]; # Simplify adjustment for possible flip;
	$ends[2] += $hgt; # Adjust height.
	$ends[5] += $hgt;
	# Track lowest Z of XYZ to know if above ground.
	if ($lowest_z > $ends[2]) { $lowest_z = $ends[2] }
	if ($lowest_z > $ends[5]) { $lowest_z = $ends[5] }
	$nec .= qq{\nGW $cnt $seg } . join(' ', @ends, $dia/2);
	$ant .= qq{\n$cnt } . join(' ', @ends[0,1,2]) . ' ' . join(' ', @ends[3,4,5], $dia);	
	$rhino .= qq{\n_Line } . join(',', @ends[0,1,2]) . ' ' . join(',', @ends[3,4,5]);
	if ($aref == \@main) {
		@end_1 = @ends[3,4,5];			
	} else {
		@end_2 = @ends[3,4,5];			
	}	
}

sub build_spiral_arm {
	my $aref = shift;
	++$cnt;
	if ($flip) {
		write_spiral($aref,2,0,1,5,3,4); # Swap X and Z.		
	} else {
		write_spiral($aref,0,1,2,3,4,5);		
	}		

	for (0,1,2) {shift @$aref}
}

sub finish_ant {	
	my @feed = feed();
	stub();
	$nec .= qq{\nGE};
	$nec .= qq{\nFR 0 }; # Starting MHz for SWR
	$nec .= int(($f_hi-$f_lo)/0.25); # How many SWR hops
	$nec .= qq{ 0 0 $f_lo 0.25}; # Size of SWR hops
	$nec .= qq{\nEX 0 1 $feed[1] 0 1 0};
	$nec .= qq{\nEN};
	my $lgth_tot = 2 * $lgth_arm;
	my $info = sprintf("\nWire each main arm = %.3f meters.", $lgth_arm);
	if ($fat> 1) {		
		$info .= sprintf("\nWire each fat arm = %.3f meters.", $lgth_sib);
		$lgth_tot += 2 * $lgth_sib;
	}
	$info .= sprintf("\nTotal wire all arms = %.3f meters.", $lgth_tot);
	$info .= sprintf("\nLowest wire = %.3f meters above ground.", $lowest_z);
	$ant .= "\nComments:$info"; 
	return $info;
}

# So that iterated filenames sort consecutively.
sub compose_file_name {
	my ($s, $r) = @_; 
	my $ant_type = 'Log_Spiral';
	if ($tau == 1) {$ant_type = 'Archemedes_Spiral'}
	if ($zee != 0) {$ant_type = "Conical_$ant_type"}
	if ($s ne '' && $r =~ m/[0-9]/) {
		# Make pretty as integer
		$s .= '_' . sprintf('%0.3f', $r);
		return "$path/$ant_type" . "_$s"; 
	} else {
		return "$path/$ant_type" ;
	}
}

sub write_ant {
	my ($s, $r, $info) = @_;
	my $file_path = compose_file_name($s, $r);
	print "\n\nWriting file: '$file_path' $info";
	open my $fh, '>', "$file_path.nec" or die ("\nOops! Cannot write NEC file '$file_path.nec': $!\n");
	print $fh $nec;
	close $fh;	
	open $fh, '>', "$file_path.ant" or die ("\nOops! Cannot write ANT file '/$file_path/.ant': $!\n");
	print $fh $ant;
	close $fh;	
	open $fh, '>', "$file_path.txt" or die ("\nOops! Cannot write TXT file '$file_path.txt': $!\n");
	print $fh $rhino;
	close $fh;	
}

sub make_ant {
	my ($s, $r, $f_md) = @_; # Scalar $f_md is NOT redunant, since can iterate for plotting.
	init_ant($s, $r, $f_md);
	my $j = 2000; # Endless loop breakout limit.
	# Spiral arm 1 of 2.
	# Continue until radius > lambda/4 and leg > lambda.
	while ($rho < $max) {
		wire();
		--$j; # Normal decrement.
		if ($j < 0) {
			die("\nExit! Infinite loop detected.");
		}
	}
	
	build_feed_wire(0,1,2);
	
	# Currently working only non-flipped and $hgt=0
	if ($fat > 1){
	fatten_feed_wire();
		@main[0,1,2] = @feed[0,1,2];
		@anti[0,1,2] = @feed[3,4,5];
		unshift(@main_sib, @feed[0,1,2]);
		unshift(@anti_sib, @feed[3,4,5]);
		# Combine 4 wires into 2 arms by joining outer tips.
		@main[-1,-2,-3] = @main_sib[-1,-2,-3];
		@anti[-1,-2,-3] = @anti_sib[-1,-2,-3];		
	}
	
	# Manage different orientations: Spin in plane;sub Flip into vertical.
	if ($flip) {
		write_feed_wire(1,0,2,4,3,5); # Swap X and Y for vertical flip.
	} else {		
		write_feed_wire(0,1,2,3,4,5);
	}
		
	# Create pair of wires for 2 arms.
	while ($#main > 2) { build_spiral_arm(\@main) } 
	while ($#anti > 2) { build_spiral_arm(\@anti) }
	# Add another pair of wires to 'broaden' the arms.
	# Geometry looks good at $fat=3, thin if $fat=2.
	# Impedance worse as a rule unless $fat=1.
	if ($fat > 1) {
		print "\nWarn: AGT will be poor on fattened spirals.";
		while ($#main_sib > 2) { build_spiral_arm(\@main_sib) } 
		while ($#anti_sib > 2) { build_spiral_arm(\@anti_sib) } 
	}
	my $info = finish_ant();
	write_ant($s, $r, $info);
}

# Yes, I should have started out with a hash...
# Nevertheless, this works too.
sub iter_value {
	my ($param, $r) = @_;
	if ($param eq 'tau') { $tau = $r }
	if ($param eq 'hgt') { $hgt = $r }
	if ($param eq 'zee') { $zee = $r }
	if ($param eq 'deg') { $deg = $r }
	if ($param eq 'f_hi') { $f_hi = $r }
	if ($param eq 'f_md') { $f_md = $r }
	if ($param eq 'f_lo') { $f_lo = $r }
	if ($param eq 'spin') { $spin = $r }
	if ($param eq 'div') {
		$div = $r;
		if ($div % 2 != 0) { die("\nOops! --div = $div (Even-sided polygons only!)\n")}
	}
}

if ($iter){
	# Edit/replace the for-loop below for custom iterating.
	for (my $r=$iter_0th; $r<=$iter_Nth; $r+=$iter_step) {	
		#$max = $iter_start{'max'};
		$lgth_arm = $lgth_sib = 0;
		iter_value($iter, $r);
		$rho = $iter_start{'rho'};
		$zee = $iter_start{'zee'} / $div;
		$zee_sib = $iter_start{'zee_sib'} / $div; 
		if ($iter eq 'div') { $tau = $tau_360**(1/$div) } blab('tau', $tau);
		make_ant($iter, $r, $f_md); # string, value, MHz
	}
} else {
	make_ant('', 0, $f_md); # string, value, MHz
}

__END__

perl KY8D_Spiral_Calculator.pl

Output 1 planar log-spiral

cls && perl KY8D_Spiral_Calculator.pl
--path r: --deg 0 --div 8 --dia #12 --f_hi 40 --f_md 14 --f_lo 13
--zee 0 --hgt 10 --seg 60 --fat 0 --tau 1.1 --flip

Output 1 conical log-spiral

cls && perl KY8D_Spiral_Calculator.pl --path r: --deg 60 --div 6 --dia #12 --f_hi 40 --f_md 14 --f_lo 13 --zee .5 --hgt 10 --seg 60 --fat 0 --tau 2.2


Iterate tau

cls && perl KY8D_Spiral_Calculator.pl --path r: --deg 0 --div 8 --dia 1.628 --f_hi 50 --f_md 24.89 --f_lo 13.5 --zee 0 --hgt 10 --seg 120 --flip --iter tau --0th 1.5 --nth 4.2 --stp 0.05


